; Pong by David Phillips <david@acz.org>
; Example source code for GBasm/GBlink

#include "gbc.inc"

;============================================================================
VBlank			= $ff83
StackTop		= _BANKV - 2
Sprites			= _BANK0 + $100	; [160]

Vars	 		= _BANK0
KeyP 			= Vars			; [1]
KeyC			= Vars + 1		; [1]
TempStr			= Vars + 2		; [6]
Paddle			= Vars + 8		; [1]
BallY			= Vars + 9		; [2]
BallX			= Vars + 11		; [2]
BallA			= Vars + 13		; [1]
Score			= Vars + 14		; [2]
Ready			= Vars + 16		; [1]

.ext DisplayACZLogo

Main:
 ld	hl,StackTop
 ld sp,hl				; create stack

 cp $11					; check if CGB unit
 jr nz,StartLevel

 ld a,1					; show logo with white background
 call DisplayACZLogo

StartLevel:
 call Setup

 sub a
 ld (KeyP),a

 ld hl,Score
 ld a,$ff
 ldi (hl),a
 ldi (hl),a
 ld hl,Paddle
 ld a,72
 ldi (hl),a
 call UpdateScore

MainLoop:
 ld hl,BallY
 sub a			; bally = 80
 ldi (hl),a
 ld a,80
 ldi (hl),a
 sub a			; ballx = 80
 ldi (hl),a
 ld a,80
 ldi (hl),a
 ld a,32		; angle = 32
 ld (hl),a

 ld hl,ReadyText
 ld de,_BG1+(8*32)+7
 call Puts
 ld a,90
 ld (Ready),a

PlayLoop:
 call DrawSprites
 call UpdatePaddle
 ld hl,Ready
 sub a
 or (hl)
 jr z,@skipready
 dec (hl)
 jr nz,@readydone
 ld hl,_BG1+(8*32)+7
 ld bc,6
 call _clearmem
@skipready:
 call UpdateBall
@readydone:

 ld a,(KeyC)
 rla
 call c,PauseGame  

 halt
 jr PlayLoop

@loop:
 halt
 jr @loop

PauseGame:
 ld hl,PauseText
 ld de,_BG1+(8*32)+7
 call Puts
@release:
 call GetKey
 ld a,(KeyP)
 rla
 jr c,@release
@loop:
 halt
 call GetKey
 ld a,(KeyC)
 rla
 jr nc,@loop
@wait:
 call GetKey
 ld a,(KeyP)
 rla
 jr c,@wait
 call WaitForVBlank
 ld hl,_BG1+(8*32)+7
 ld bc,6
 call _clearmem
 ret  

UpdateBall:
 ld a,(BallA)
 call LoadSine16
 ld hl,BallY
 call AddAddress
 call AddAddress 
 ld a,(BallA)
 add a,64
 call LoadSine16 
 ld hl,BallX
 call AddAddress
 call AddAddress 
 ld hl,BallA
 ld a,(BallX+1)
 or a
 jr nz,@notleft
 ld b,(hl)
 ld a,128
 sub b
 ld (hl),a
 ld a,$ff
 ld (BallX),a
 ret
@notleft:
 cp 156
 jr nz,@notright
 ld b,(hl)
 ld a,128
 sub b
 ld (hl),a
 sub a
 ld (BallX),a
 ret
@notright:
 ld a,(BallY+1)
 cp 8
 jr nz,@nottop
 ld b,(hl)
 sub a
 sub b
 ld (hl),a
 ld a,$ff
 ld (BallY),a
 jr UpdateScore
@nottop:
 cp 144
 jr nz,@notbottom
 pop hl
 jp MainLoop
@notbottom: 
 cp 124
 ret c
 cp 128
 ret nc
 ld a,(Paddle)
 dec a
 ld b,a
 ld a,(BallX+1)
 sub b
 cp 18
 ret nc
 ld b,(hl)
 sub a
 sub b
 ld (hl),a
 sub a
 ld (BallY),a
 ret

LoadSine16:
 ld d,0
 ld e,a
 ld hl,SineTable
 add hl,de
 ld a,(hl)
 ld e,a
 rla
 jr nc,@notneg
 dec d
 res 7,e
@notneg:
 ret

AddAddress:
 push de
 push hl
 ldi a,(hl)
 ld h,(hl)
 ld l,a
 add hl,de
 push hl
 pop de
 pop hl
 ld (hl),e
 inc hl
 ld (hl),d
 dec hl
 pop de
 ret

UpdateScore:
 ld hl,Score
 ld d,(hl)
 inc hl
 ld e,(hl)
 inc de
 ld (hl),e
 dec hl
 ld (hl),d
 ld hl,_BG1+15
 call DispDE
 ret

UpdatePaddle:
 call GetKey
 ld hl,Paddle
 ld a,(KeyP)
 rra
 jr c,@right
 rra
 jr c,@left
 ret
@right:
 ld a,(hl)
 cp 143
 ret nc
 inc (hl)
 inc (hl)
 ret
@left:
 ld a,(hl)
 cp 2
 ret c
 dec (hl)
 dec (hl)
 ret

DrawSprites:
 ld hl,Sprites
 ld a,144
 ldi (hl),a
 ld a,(Paddle)
 add a,8
 ldi (hl),a
 sub a
 ldi (hl),a
 ldi (hl),a
 ld a,144
 ldi (hl),a
 ld a,(Paddle)
 add a,16
 ldi (hl),a
 ld a,1
 ldi (hl),a
 sub a
 ldi (hl),a
 ld a,(BallY+1)
 add a,16
 ldi (hl),a
 ld a,(BallX+1)
 add a,8
 ldi (hl),a
 ld a,2
 ldi (hl),a
 sub a
 ldi (hl),a
 ret

PauseText:
 .db "PAUSED",0

ReadyText:
 .db "Ready?",0

;============================================================================
Setup:
 call WaitForVBlank		; wait for vblank to turn off lcd

 sub a					; disable everything, lcd operation stopped
 ldh (LCDC),a

; sub a					; set background to cover screen
 ldh (SCY),a
 ldh (SCX),a

; sub a					; swap in bank 0, copy font sprites
 ldh (VBK),a
 ld hl,FontChars
 ld de,_WNPA
 ld bc,$60*16
 call _ldir

 ld hl,ToplineSprite	; copy sprite for horizontal bar
 ld de,_WNPA+($60*16)
 ld bc,16
 call _ldir

 ld hl,PaddleSprites	; copy paddle and ball sprites to object memory
 ld de,_OBPA
 ld bc,16*3
 call _ldir
 
 ld hl,_BG1			; clear window tile memory
 ld bc,$400
 call _clearmem

 ld a,1					; clear window attribute memory
 ldh (VBK),a
 ld hl,_BG1
 ld bc,$400
 call _clearmem
 ldh (VBK),a

 ld hl,Sprites			; clear sprite memory
 ld bc,4*40
 call _clearmem

 ld hl,VBlankIntr		; copy DMA handler to high ram
 ld de,VBlank
 ld bc,VBlankIntrEnd-VBlankIntr
 call _ldir

 ld hl,_BG1+32			; draw horizontal bar
 ld bc,20
 ld a,$60
 call _setram

 ld hl,GrayPalette		; set bg/window palette 0 to gray
 ld b,8
 call LoadBGPalettes

 ld hl,GrayPalette		; set object palette 0 to gray
 ld b,8
 call LoadOPalettes

 ld a,%11100100			; set DMG palettes to grayscale
 ldh (BGP),a			; set bg/window
 ldh (OBP0),a			; set object palette 0

 ld a,%00000001			; enable vblank, disable all other interrupts
 ldh (IE),a
 ei

 ld a,%11000011			; use _BG1, window off, 8x8 sprites on, bg on
 ldh (LCDC),a 

 ld hl,StatusText		; draw status line
 ld de,_BG1
 call Puts

 ret

VBlankIntr:
 push af
 ld a,Sprites>>8
 ldh (DMA),a
 ld a,$28
@wait:
 dec a
 jr nz,@wait
 pop af
 ret
VBlankIntrEnd

StatusText:
.db "Pong!   Score:",0

ToplineSprite:
 .dw $ffff,$0000,$0000,$0000
 .dw $0000,$0000,$0000,$0000

GrayPalette:
 .dw %0111111111111111	; R:31,G:31,B:31
 .dw %0101001010010100	; R:20,G:20,B:20
 .dw %0010100101001010	; R:10,G:10,B:10
 .dw %0000000000000000	; R:0,G:0,B:0

PaddleSprites:
 .db %01111111
 .db %01111111
 .db %11111111
 .db %10000000
 .db %11111111
 .db %10000000
 .db %01111111
 .db %01111111
 .dw 0,0,0,0

 .db %11111110
 .db %11111110
 .db %11111111
 .db %00000001
 .db %11111111
 .db %00000001
 .db %11111110
 .db %11111110
 .dw 0,0,0,0

BallSprite:
 .db %01100000
 .db %01100000
 .db %11110000
 .db %10010000
 .db %11110000
 .db %10010000
 .db %01100000
 .db %01100000
 .dw 0,0,0,0


;============================================================================
; hl = source, de = dest, bc = bytes
_ldir:
 ldi a,(hl)
 ld (de),a
 inc de
 dec bc
 ld a,b
 or c
 jr nz,_ldir
 ret

; hl = dest, bc = bytes
_clearmem:
 sub a
 ldi (hl),a
 dec bc
 ld a,b
 or c
 jr nz,_clearmem
 ret

; hl = dest, bc = bytes, a = char
_setram:
 push de
 ld d,h
 ld e,l
@loop:
 ldi (hl),a
 dec bc
 ld a,b
 or c
 ld a,(de)
 jr nz,@loop
 pop de
 ret 

WaitForVBlank:
 ldh a,(LY)
 cp 145
 jr nz,WaitForVBlank
 ret

WaitForHBlank:
 ldh a,(STAT)
 and %00000011
 jr nz,WaitForHBlank
 ret

;*************************************************
;GET KEYPRESSES
;keypresses stored in (KeyP)
;change in keypresses stored in (KeyC)
;bit 7 - start
;bit 6 - select
;bit 5 - B
;bit 4 - A
;bit 3 - down
;bit 2 - up
;bit 1 - left
;bit 0 - right
;*************************************************
GetKey:						;gets keypresses
 ld a,P15
 ldh (P1),a					;set P15
 nop \ nop \ nop
 nop \ nop \ nop
 ldh a,(P1)					;read control keys
 and %1111
 swap a
 ld b,a						;save those four bits for later
 ld a,P14
 ldh (P1),a					;set P15
 nop \ nop \ nop
 nop \ nop \ nop
 nop \ nop \ nop
 nop \ nop \ nop
 ldh a,(P1)					;read arrow keys
 and %1111					;keep the keypresses
 or b						;add control key data
 cpl						;make bits set if keypress
 ld b,a						;put it back in b
 ld a,P0
 ldh (P1),a					;deselect ports
 ld hl,KeyP
 ldi a,(hl)					;get old keypress
 xor b						;find changes
 ldd (hl),a					;store changes
 ld (hl),b					;store current keypress
 ld a,b						;handle 4-button reset
 and %11110000
 cp %11110000
 jp z,$150  
 ret

;*************************************************
;UNPACK DE
;return a string at HL from a value in DE
;*************************************************
UnpackDE:
 ld (hl),0
 dec hl
@loop:
 ld c,10
 call @divide
 add a,'0'
 ldd (hl),a					;store remainder into string
 ld a,d
 or e						;test DE
 jr nz,@loop
 inc hl
 ret
@divide:
 xor a
 ld b,16		;16 shifts
@dloop:
 sla e
 rl d 			;shift DE left
 rla 			;into A
 cp c 			;if A<=C
 jr c,@skip
 sub c			;then A= A-C
 inc e			;set bit 0
@skip:
 dec b
 jr nz,@dloop
 ret

;input= h and l
;output= de is h*l (all reg destroyed but c)
MultHL:
 ld de,0		;clear de initially
 ld b,h
 ld h,d
@loop:
 srl b			;if bit 0 is set, add hl to de
 jr nc,@skipCollect
 push hl
 add hl,de
 ld d,h
 ld e,l
 pop hl
@skipCollect:
 ret z			;return when b is zero (still set from srl)
 add hl,hl		;shift hl left
 jr @loop


;*************************************************
;Set B bytes of palette data at HL into palettes
;*************************************************
LoadBGPalettes:
 ld a,%10000000				;select auto increment and palette 0, color 0
 ldh (BCPS),a
@loop:
 ldi a,(hl)
 ldh (BCPD),a				;copy palette data from HL
 dec b						;B times
 jr nz,@loop
 ret

LoadOPalettes:
 ld a,%10000000				;select auto increment and palette 0, color 0
 ldh (OCPS),a
@loop:
 ldi a,(hl)
 ldh (OCPD),a				;copy palette data from HL
 dec b						;B times
 jr nz,@loop
 ret

; de = number, hl = tile memory position
DispDE:
 push hl
 ld hl,TempStr+5
 call UnpackDE
 pop de
;*************************************************
;Copy HL to DE until first null
;adjust sprite numbers for text
;*************************************************
Puts:
 call WaitForHBlank
 ldi a,(hl)
 or a
 ret z
 sub $20
 ld (de),a
 inc de
 jr Puts

;============================================================================
SineTable:
 .db   0,   3,   6,   9,  12,  15,  18,  21,  24,  27
 .db  30,  33,  36,  39,  42,  45,  48,  51,  53,  56
 .db  59,  62,  64,  67,  70,  72,  75,  77,  79,  82
 .db  84,  86,  89,  91,  93,  95,  97,  99, 101, 103
 .db 104, 106, 108, 109, 111, 112, 113, 115, 116, 117
 .db 118, 119, 120, 121, 122, 122, 123, 124, 124, 125
 .db 125, 125, 125, 126, 126, 126, 125, 125, 125, 125
 .db 124, 124, 123, 122, 122, 121, 120, 119, 118, 117
 .db 116, 115, 113, 112, 111, 109, 108, 106, 104, 103
 .db 101,  99,  97,  95,  93,  91,  89,  86,  84,  82
 .db  79,  77,  75,  72,  70,  67,  64,  62,  59,  56
 .db  53,  51,  48,  45,  42,  39,  36,  33,  30,  27
 .db  24,  21,  18,  15,  12,   9,   6,   3,   0, 253
 .db 250, 247, 244, 241, 238, 235, 232, 229, 226, 223
 .db 220, 217, 214, 211, 208, 205, 203, 200, 197, 194
 .db 192, 189, 186, 184, 181, 179, 177, 174, 172, 170
 .db 167, 165, 163, 161, 159, 157, 155, 153, 152, 150
 .db 148, 147, 145, 144, 143, 141, 140, 139, 138, 137
 .db 136, 135, 134, 134, 133, 132, 132, 131, 131, 131
 .db 131, 130, 130, 130, 131, 131, 131, 131, 132, 132
 .db 133, 134, 134, 135, 136, 137, 138, 139, 140, 141
 .db 143, 144, 145, 147, 148, 150, 152, 153, 155, 157
 .db 159, 161, 163, 165, 167, 170, 172, 174, 177, 179
 .db 181, 184, 186, 189, 192, 194, 197, 200, 203, 205
 .db 208, 211, 214, 217, 220, 223, 226, 229, 232, 235
 .db 238, 241, 244, 247, 250, 253

#include "charfont.asm"

.end
